# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for PauliList class."""

import itertools
import unittest

import numpy as np
from ddt import ddt
from scipy.sparse import csr_matrix

from qiskit import QiskitError
from qiskit.circuit.library import (
    CXGate,
    CYGate,
    CZGate,
    HGate,
    IGate,
    SdgGate,
    SGate,
    SwapGate,
    XGate,
    YGate,
    ZGate,
)
from qiskit.quantum_info.operators import (
    Clifford,
    Operator,
    Pauli,
    PauliList,
)
from qiskit.quantum_info.random import random_clifford, random_pauli_list
from test import combine  # pylint: disable=wrong-import-order
from test import QiskitTestCase  # pylint: disable=wrong-import-order

from .test_pauli import pauli_group_labels


def pauli_mat(label):
    """Return Pauli matrix from a Pauli label"""
    mat = np.eye(1, dtype=complex)
    if label[0:2] == "-i":
        mat *= -1j
        label = label[2:]
    elif label[0] == "-":
        mat *= -1
        label = label[1:]
    elif label[0] == "i":
        mat *= 1j
        label = label[1:]
    for i in label:
        if i == "I":
            mat = np.kron(mat, np.eye(2, dtype=complex))
        elif i == "X":
            mat = np.kron(mat, np.array([[0, 1], [1, 0]], dtype=complex))
        elif i == "Y":
            mat = np.kron(mat, np.array([[0, -1j], [1j, 0]], dtype=complex))
        elif i == "Z":
            mat = np.kron(mat, np.array([[1, 0], [0, -1]], dtype=complex))
        else:
            raise QiskitError(f"Invalid Pauli string {i}")
    return mat


class TestPauliListInit(QiskitTestCase):
    """Tests for PauliList initialization."""

    def test_array_init(self):
        """Test array initialization."""
        # Matrix array initialization

        with self.subTest(msg="Empty array"):
            x = np.array([], dtype=bool).reshape((1, 0))
            z = np.array([], dtype=bool).reshape((1, 0))
            pauli_list = PauliList.from_symplectic(x, z)
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg="bool array"):
            z = np.array([[False], [True]])
            x = np.array([[False], [True]])
            pauli_list = PauliList.from_symplectic(z, x)
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg="bool array no copy"):
            z = np.array([[False], [True]])
            x = np.array([[True], [True]])
            pauli_list = PauliList.from_symplectic(z, x)
            z[0, 0] = not z[0, 0]
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

    def test_string_init(self):
        """Test string initialization."""
        # String initialization
        with self.subTest(msg='str init "I"'):
            pauli_list = PauliList("I")
            z = np.array([[False]])
            x = np.array([[False]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg='str init "X"'):
            pauli_list = PauliList("X")
            z = np.array([[False]])
            x = np.array([[True]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg='str init "Y"'):
            pauli_list = PauliList("Y")
            z = np.array([[True]])
            x = np.array([[True]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg='str init "Z"'):
            pauli_list = PauliList("Z")
            z = np.array([[True]])
            x = np.array([[False]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg='str init "iZ"'):
            pauli_list = PauliList("iZ")
            z = np.array([[True]])
            x = np.array([[False]])
            phase = np.array([3])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)
            np.testing.assert_equal(pauli_list.phase, phase)

        with self.subTest(msg='str init "-Z"'):
            pauli_list = PauliList("-Z")
            z = np.array([[True]])
            x = np.array([[False]])
            phase = np.array([2])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)
            np.testing.assert_equal(pauli_list.phase, phase)

        with self.subTest(msg='str init "-iZ"'):
            pauli_list = PauliList("-iZ")
            z = np.array([[True]])
            x = np.array([[False]])
            phase = np.array([1])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)
            np.testing.assert_equal(pauli_list.phase, phase)

        with self.subTest(msg='str init "IX"'):
            pauli_list = PauliList("IX")
            z = np.array([[False, False]])
            x = np.array([[True, False]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg='str init "XI"'):
            pauli_list = PauliList("XI")
            z = np.array([[False, False]])
            x = np.array([[False, True]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg='str init "YZ"'):
            pauli_list = PauliList("YZ")
            z = np.array([[True, True]])
            x = np.array([[False, True]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg='str init "iZY"'):
            pauli_list = PauliList("iZY")
            z = np.array([[True, True]])
            x = np.array([[True, False]])
            phase = np.array([3])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)
            np.testing.assert_equal(pauli_list.phase, phase)

        with self.subTest(msg='str init "XIZ"'):
            pauli_list = PauliList("XIZ")
            z = np.array([[True, False, False]])
            x = np.array([[False, False, True]])
            np.testing.assert_equal(pauli_list.z, z)
            np.testing.assert_equal(pauli_list.x, x)

        with self.subTest(msg="str init prevent broadcasting"):
            with self.assertRaises(ValueError):
                PauliList(["XYZ", "I"])

    def test_list_init(self):
        """Test list initialization."""
        with self.subTest(msg="PauliList"):
            target = PauliList(["iXI", "IX", "IZ"])
            value = PauliList(target)
            self.assertEqual(value, target)

        with self.subTest(msg="PauliList no copy"):
            target = PauliList(["iXI", "IX", "IZ"])
            value = PauliList(target)
            value[0] = "-iII"
            self.assertEqual(value, target)

    def test_init_from_settings(self):
        """Test initializing from the settings dictionary."""
        pauli_list = PauliList(["IX", "-iYZ", "YY"])
        from_settings = PauliList(**pauli_list.settings)
        self.assertEqual(pauli_list, from_settings)


@ddt
class TestPauliListProperties(QiskitTestCase):
    """Tests for PauliList properties."""

    def test_x_property(self):
        """Test X property"""
        with self.subTest(msg="X"):
            pauli = PauliList(["XI", "IZ", "YY"])
            array = np.array([[False, True], [False, False], [True, True]], dtype=bool)
            self.assertTrue(np.all(pauli.x == array))

        with self.subTest(msg="set X"):
            pauli = PauliList(["XI", "IZ"])
            val = np.array([[False, False], [True, True]], dtype=bool)
            pauli.x = val
            self.assertEqual(pauli, PauliList(["II", "iXY"]))

        with self.subTest(msg="set X raises"):
            with self.assertRaises(Exception):
                pauli = PauliList(["XI", "IZ"])
                val = np.array([[False, False, False], [True, True, True]], dtype=bool)
                pauli.x = val

    def test_z_property(self):
        """Test Z property"""
        with self.subTest(msg="Z"):
            pauli = PauliList(["XI", "IZ", "YY"])
            array = np.array([[False, False], [True, False], [True, True]], dtype=bool)
            self.assertTrue(np.all(pauli.z == array))

        with self.subTest(msg="set Z"):
            pauli = PauliList(["XI", "IZ"])
            val = np.array([[False, False], [True, True]], dtype=bool)
            pauli.z = val
            self.assertEqual(pauli, PauliList(["XI", "ZZ"]))

        with self.subTest(msg="set Z raises"):
            with self.assertRaises(Exception):
                pauli = PauliList(["XI", "IZ"])
                val = np.array([[False, False, False], [True, True, True]], dtype=bool)
                pauli.z = val

    def test_phase_property(self):
        """Test phase property"""
        with self.subTest(msg="phase"):
            pauli = PauliList(["XI", "IZ", "YY", "YI"])
            array = np.array([0, 0, 0, 0], dtype=int)
            np.testing.assert_equal(pauli.phase, array)

        with self.subTest(msg="set phase"):
            pauli = PauliList(["XI", "IZ"])
            val = np.array([2, 3], dtype=int)
            pauli.phase = val
            self.assertEqual(pauli, PauliList(["-XI", "iIZ"]))

        with self.subTest(msg="set Z raises"):
            with self.assertRaises(Exception):
                pauli = PauliList(["XI", "IZ"])
                val = np.array([1, 2, 3], dtype=int)
                pauli.phase = val

    def test_shape_property(self):
        """Test shape property"""
        shape = (3, 4)
        pauli = PauliList.from_symplectic(np.zeros(shape), np.zeros(shape))
        self.assertEqual(pauli.shape, shape)

    @combine(j=range(1, 10))
    def test_size_property(self, j):
        """Test size property"""
        shape = (j, 4)
        pauli = PauliList.from_symplectic(np.zeros(shape), np.zeros(shape))
        self.assertEqual(len(pauli), j)

    @combine(j=range(1, 10))
    def test_n_qubit_property(self, j):
        """Test n_qubit property"""
        shape = (5, j)
        pauli = PauliList.from_symplectic(np.zeros(shape), np.zeros(shape))
        self.assertEqual(pauli.num_qubits, j)

    def test_eq(self):
        """Test __eq__ method."""
        pauli1 = PauliList(["II", "XI"])
        pauli2 = PauliList(["XI", "II"])
        self.assertEqual(pauli1, pauli1)
        self.assertNotEqual(pauli1, pauli2)

    def test_len_methods(self):
        """Test __len__ method."""
        for j in range(1, 10):
            labels = j * ["XX"]
            pauli = PauliList(labels)
            self.assertEqual(len(pauli), j)

    def test_add_methods(self):
        """Test __add__ method."""
        labels1 = ["XXI", "IXX"]
        labels2 = ["XXI", "ZZI", "ZYZ"]
        pauli1 = PauliList(labels1)
        pauli2 = PauliList(labels2)
        target = PauliList(labels1 + labels2)
        self.assertEqual(target, pauli1 + pauli2)

    def test_add_qargs(self):
        """Test add method with qargs."""
        pauli1 = PauliList(["IIII", "YYYY"])
        pauli2 = PauliList(["XY", "YZ"])
        pauli3 = PauliList(["X", "Y", "Z"])

        with self.subTest(msg="qargs=[0, 1]"):
            target = PauliList(["IIII", "YYYY", "IIXY", "IIYZ"])
            self.assertEqual(pauli1 + pauli2([0, 1]), target)

        with self.subTest(msg="qargs=[0, 3]"):
            target = PauliList(["IIII", "YYYY", "XIIY", "YIIZ"])
            self.assertEqual(pauli1 + pauli2([0, 3]), target)

        with self.subTest(msg="qargs=[2, 1]"):
            target = PauliList(["IIII", "YYYY", "IYXI", "IZYI"])
            self.assertEqual(pauli1 + pauli2([2, 1]), target)

        with self.subTest(msg="qargs=[3, 1]"):
            target = PauliList(["IIII", "YYYY", "YIXI", "ZIYI"])
            self.assertEqual(pauli1 + pauli2([3, 1]), target)

        with self.subTest(msg="qargs=[0]"):
            target = PauliList(["IIII", "YYYY", "IIIX", "IIIY", "IIIZ"])
            self.assertEqual(pauli1 + pauli3([0]), target)

        with self.subTest(msg="qargs=[1]"):
            target = PauliList(["IIII", "YYYY", "IIXI", "IIYI", "IIZI"])
            self.assertEqual(pauli1 + pauli3([1]), target)

        with self.subTest(msg="qargs=[2]"):
            target = PauliList(["IIII", "YYYY", "IXII", "IYII", "IZII"])
            self.assertEqual(pauli1 + pauli3([2]), target)

        with self.subTest(msg="qargs=[3]"):
            target = PauliList(["IIII", "YYYY", "XIII", "YIII", "ZIII"])
            self.assertEqual(pauli1 + pauli3([3]), target)

    def test_getitem_methods(self):
        """Test __getitem__ method."""
        with self.subTest(msg="__getitem__ single"):
            labels = ["XI", "IY"]
            pauli = PauliList(labels)
            self.assertEqual(pauli[0], PauliList(labels[0]))
            self.assertEqual(pauli[1], PauliList(labels[1]))

        with self.subTest(msg="__getitem__ array"):
            labels = np.array(["XI", "IY", "IZ", "XY", "ZX"])
            pauli = PauliList(labels)
            inds = [0, 3]
            self.assertEqual(pauli[inds], PauliList(labels[inds]))
            inds = np.array([4, 1])
            self.assertEqual(pauli[inds], PauliList(labels[inds]))

        with self.subTest(msg="__getitem__ slice"):
            labels = np.array(["XI", "IY", "IZ", "XY", "ZX"])
            pauli = PauliList(labels)
            self.assertEqual(pauli[:], pauli)
            self.assertEqual(pauli[1:3], PauliList(labels[1:3]))

    def test_setitem_methods(self):
        """Test __setitem__ method."""
        with self.subTest(msg="__setitem__ single"):
            labels = ["XI", "IY"]
            pauli = PauliList(["XI", "IY"])
            pauli[0] = "II"
            self.assertEqual(pauli[0], PauliList("II"))
            pauli[1] = "-iXX"
            self.assertEqual(pauli[1], PauliList("-iXX"))

            with self.assertRaises(Exception):
                # Wrong size Pauli
                pauli[0] = "XXX"

        with self.subTest(msg="__setitem__ array"):
            labels = np.array(["XI", "IY", "IZ"])
            pauli = PauliList(labels)
            target = PauliList(["II", "ZZ"])
            inds = [2, 0]
            pauli[inds] = target
            self.assertEqual(pauli[inds], target)

            with self.assertRaises(Exception):
                pauli[inds] = PauliList(["YY", "ZZ", "XX"])

        with self.subTest(msg="__setitem__ slice"):
            labels = np.array(5 * ["III"])
            pauli = PauliList(labels)
            target = PauliList(5 * ["XXX"])
            pauli[:] = target
            self.assertEqual(pauli[:], target)
            target = PauliList(2 * ["ZZZ"])
            pauli[1:3] = target
            self.assertEqual(pauli[1:3], target)


class TestPauliListLabels(QiskitTestCase):
    """Tests PauliList label representation conversions."""

    def test_from_labels_1q(self):
        """Test 1-qubit from_labels method."""
        labels = ["I", "Z", "Z", "X", "Y"]
        target = PauliList.from_symplectic(
            np.array([[False], [True], [True], [False], [True]]),
            np.array([[False], [False], [False], [True], [True]]),
        )
        value = PauliList(labels)
        self.assertEqual(target, value)

    def test_from_labels_1q_with_phase(self):
        """Test 1-qubit from_labels method with phase."""
        labels = ["-I", "iZ", "iZ", "X", "-iY"]
        target = PauliList.from_symplectic(
            np.array([[False], [True], [True], [False], [True]]),
            np.array([[False], [False], [False], [True], [True]]),
            np.array([2, 3, 3, 0, 1]),
        )
        value = PauliList(labels)
        self.assertEqual(target, value)

    def test_from_labels_2q(self):
        """Test 2-qubit from_labels method."""
        labels = ["II", "YY", "XZ"]
        target = PauliList.from_symplectic(
            np.array([[False, False], [True, True], [True, False]]),
            np.array([[False, False], [True, True], [False, True]]),
        )
        value = PauliList(labels)
        self.assertEqual(target, value)

    def test_from_labels_2q_with_phase(self):
        """Test 2-qubit from_labels method."""
        labels = ["iII", "iYY", "-iXZ"]
        target = PauliList.from_symplectic(
            np.array([[False, False], [True, True], [True, False]]),
            np.array([[False, False], [True, True], [False, True]]),
            np.array([3, 3, 1]),
        )
        value = PauliList(labels)
        self.assertEqual(target, value)

    def test_from_labels_5q(self):
        """Test 5-qubit from_labels method."""
        labels = [5 * "I", 5 * "X", 5 * "Y", 5 * "Z"]
        target = PauliList.from_symplectic(
            np.array([[False] * 5, [False] * 5, [True] * 5, [True] * 5]),
            np.array([[False] * 5, [True] * 5, [True] * 5, [False] * 5]),
        )
        value = PauliList(labels)
        self.assertEqual(target, value)

    def test_to_labels_1q(self):
        """Test 1-qubit to_labels method."""
        pauli = PauliList.from_symplectic(
            np.array([[False], [True], [True], [False], [True]]),
            np.array([[False], [False], [False], [True], [True]]),
        )
        target = ["I", "Z", "Z", "X", "Y"]
        value = pauli.to_labels()
        self.assertEqual(value, target)

    def test_to_labels_1q_with_phase(self):
        """Test 1-qubit to_labels method with phase."""
        pauli = PauliList.from_symplectic(
            np.array([[False], [True], [True], [False], [True]]),
            np.array([[False], [False], [False], [True], [True]]),
            np.array([1, 3, 2, 3, 1]),
        )
        target = ["-iI", "iZ", "-Z", "iX", "-iY"]
        value = pauli.to_labels()
        self.assertEqual(value, target)

    def test_to_labels_1q_array(self):
        """Test 1-qubit to_labels method w/ array=True."""
        pauli = PauliList.from_symplectic(
            np.array([[False], [True], [True], [False], [True]]),
            np.array([[False], [False], [False], [True], [True]]),
        )
        target = np.array(["I", "Z", "Z", "X", "Y"])
        value = pauli.to_labels(array=True)
        self.assertTrue(np.all(value == target))

    def test_to_labels_1q_array_with_phase(self):
        """Test 1-qubit to_labels method w/ array=True."""
        pauli = PauliList.from_symplectic(
            np.array([[False], [True], [True], [False], [True]]),
            np.array([[False], [False], [False], [True], [True]]),
            np.array([2, 3, 0, 1, 0]),
        )
        target = np.array(["-I", "iZ", "Z", "-iX", "Y"])
        value = pauli.to_labels(array=True)
        self.assertTrue(np.all(value == target))

    def test_labels_round_trip(self):
        """Test from_labels and to_labels round trip."""
        target = ["III", "IXZ", "XYI", "ZZZ", "-iZIX", "-IYX"]
        value = PauliList(target).to_labels()
        self.assertEqual(value, target)

    def test_labels_round_trip_array(self):
        """Test from_labels and to_labels round trip w/ array=True."""
        labels = ["III", "IXZ", "XYI", "ZZZ", "-iZIX", "-IYX"]
        target = np.array(labels)
        value = PauliList(labels).to_labels(array=True)
        self.assertTrue(np.all(value == target))


class TestPauliListMatrix(QiskitTestCase):
    """Tests PauliList matrix representation conversions."""

    def test_to_matrix_1q(self):
        """Test 1-qubit to_matrix method."""
        labels = ["X", "I", "Z", "Y"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix()
        self.assertTrue(isinstance(values, list))
        for target, value in zip(targets, values):
            self.assertTrue(np.all(value == target))

    def test_to_matrix_1q_array(self):
        """Test 1-qubit to_matrix method w/ array=True."""
        labels = ["Z", "I", "Y", "X"]
        target = np.array([pauli_mat(i) for i in labels])
        value = PauliList(labels).to_matrix(array=True)
        self.assertTrue(isinstance(value, np.ndarray))
        self.assertTrue(np.all(value == target))

    def test_to_matrix_1q_sparse(self):
        """Test 1-qubit to_matrix method w/ sparse=True."""
        labels = ["X", "I", "Z", "Y"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix(sparse=True)
        for mat, targ in zip(values, targets):
            self.assertTrue(isinstance(mat, csr_matrix))
            self.assertTrue(np.all(targ == mat.toarray()))

    def test_to_matrix_2q(self):
        """Test 2-qubit to_matrix method."""
        labels = ["IX", "YI", "II", "ZZ"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix()
        self.assertTrue(isinstance(values, list))
        for target, value in zip(targets, values):
            self.assertTrue(np.all(value == target))

    def test_to_matrix_2q_array(self):
        """Test 2-qubit to_matrix method w/ array=True."""
        labels = ["ZZ", "XY", "YX", "IZ"]
        target = np.array([pauli_mat(i) for i in labels])
        value = PauliList(labels).to_matrix(array=True)
        self.assertTrue(isinstance(value, np.ndarray))
        self.assertTrue(np.all(value == target))

    def test_to_matrix_2q_sparse(self):
        """Test 2-qubit to_matrix method w/ sparse=True."""
        labels = ["IX", "II", "ZY", "YZ"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix(sparse=True)
        for mat, targ in zip(values, targets):
            self.assertTrue(isinstance(mat, csr_matrix))
            self.assertTrue(np.all(targ == mat.toarray()))

    def test_to_matrix_5q(self):
        """Test 5-qubit to_matrix method."""
        labels = ["IXIXI", "YZIXI", "IIXYZ"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix()
        self.assertTrue(isinstance(values, list))
        for target, value in zip(targets, values):
            self.assertTrue(np.all(value == target))

    def test_to_matrix_5q_sparse(self):
        """Test 5-qubit to_matrix method w/ sparse=True."""
        labels = ["XXXYY", "IXIZY", "ZYXIX"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix(sparse=True)
        for mat, targ in zip(values, targets):
            self.assertTrue(isinstance(mat, csr_matrix))
            self.assertTrue(np.all(targ == mat.toarray()))

    def test_to_matrix_5q_with_phase(self):
        """Test 5-qubit to_matrix method with phase."""
        labels = ["iIXIXI", "-YZIXI", "-iIIXYZ"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix()
        self.assertTrue(isinstance(values, list))
        for target, value in zip(targets, values):
            self.assertTrue(np.all(value == target))

    def test_to_matrix_5q_sparse_with_phase(self):
        """Test 5-qubit to_matrix method w/ sparse=True with phase."""
        labels = ["iXXXYY", "-IXIZY", "-iZYXIX"]
        targets = [pauli_mat(i) for i in labels]
        values = PauliList(labels).to_matrix(sparse=True)
        for mat, targ in zip(values, targets):
            self.assertTrue(isinstance(mat, csr_matrix))
            self.assertTrue(np.all(targ == mat.toarray()))


class TestPauliListIteration(QiskitTestCase):
    """Tests for PauliList iterators class."""

    def test_enumerate(self):
        """Test enumerate with PauliList."""
        labels = ["III", "IXI", "IYY", "YIZ", "XYZ", "III"]
        pauli = PauliList(labels)
        for idx, i in enumerate(pauli):
            self.assertEqual(i, PauliList(labels[idx]))

    def test_iter(self):
        """Test iter with PauliList."""
        labels = ["III", "IXI", "IYY", "YIZ", "XYZ", "III"]
        pauli = PauliList(labels)
        for idx, i in enumerate(iter(pauli)):
            self.assertEqual(i, PauliList(labels[idx]))

    def test_zip(self):
        """Test zip with PauliList."""
        labels = ["III", "IXI", "IYY", "YIZ", "XYZ", "III"]
        pauli = PauliList(labels)
        for label, i in zip(labels, pauli):
            self.assertEqual(i, PauliList(label))

    def test_label_iter(self):
        """Test PauliList label_iter method."""
        labels = ["III", "IXI", "IYY", "YIZ", "XYZ", "III"]
        pauli = PauliList(labels)
        for idx, i in enumerate(pauli.label_iter()):
            self.assertEqual(i, labels[idx])

    def test_matrix_iter(self):
        """Test PauliList dense matrix_iter method."""
        labels = ["III", "IXI", "IYY", "YIZ", "XYZ", "III"]
        pauli = PauliList(labels)
        for idx, i in enumerate(pauli.matrix_iter()):
            self.assertTrue(np.all(i == pauli_mat(labels[idx])))

    def test_matrix_iter_sparse(self):
        """Test PauliList sparse matrix_iter method."""
        labels = ["III", "IXI", "IYY", "YIZ", "XYZ", "III"]
        pauli = PauliList(labels)
        for idx, i in enumerate(pauli.matrix_iter(sparse=True)):
            self.assertTrue(isinstance(i, csr_matrix))
            self.assertTrue(np.all(i.toarray() == pauli_mat(labels[idx])))


@ddt
class TestPauliListOperator(QiskitTestCase):
    """Tests for PauliList base operator methods."""

    @combine(j=range(1, 10))
    def test_tensor(self, j):
        """Test tensor method j={j}."""
        labels1 = ["XX", "YY"]
        labels2 = [j * "I", j * "Z"]
        pauli1 = PauliList(labels1)
        pauli2 = PauliList(labels2)

        value = pauli1.tensor(pauli2)
        target = PauliList([l1 + l2 for l1 in labels1 for l2 in labels2])
        self.assertEqual(value, target)

    @combine(j=range(1, 10))
    def test_tensor_with_phase(self, j):
        """Test tensor method j={j} with phase."""
        labels1 = ["XX", "iYY"]
        labels2 = [j * "I", "i" + j * "Z"]
        pauli1 = PauliList(labels1)
        pauli2 = PauliList(labels2)

        value = pauli1.tensor(pauli2)
        target = PauliList(["XX" + "I" * j, "iXX" + "Z" * j, "iYY" + "I" * j, "-YY" + "Z" * j])
        self.assertEqual(value, target)

    @combine(j=range(1, 10))
    def test_expand(self, j):
        """Test expand method j={j}."""
        labels1 = ["XX", "YY"]
        labels2 = [j * "I", j * "Z"]
        pauli1 = PauliList(labels1)
        pauli2 = PauliList(labels2)

        value = pauli1.expand(pauli2)
        target = PauliList([j + i for j in labels2 for i in labels1])
        self.assertEqual(value, target)

    @combine(j=range(1, 10))
    def test_expand_with_phase(self, j):
        """Test expand method j={j}."""
        labels1 = ["-XX", "iYY"]
        labels2 = ["i" + j * "I", "-i" + j * "Z"]
        pauli1 = PauliList(labels1)
        pauli2 = PauliList(labels2)

        value = pauli1.expand(pauli2)
        target = PauliList(
            ["-i" + "I" * j + "XX", "-" + "I" * j + "YY", "i" + "Z" * j + "XX", "Z" * j + "YY"]
        )
        self.assertEqual(value, target)

    def test_compose_1q(self):
        """Test 1-qubit compose methods."""
        # Test single qubit Pauli dot products
        pauli = PauliList(["I", "X", "Y", "Z"])

        with self.subTest(msg="compose single I"):
            target = PauliList(["I", "X", "Y", "Z"])
            value = pauli.compose("I")
            self.assertEqual(target, value)

        with self.subTest(msg="compose single X"):
            target = PauliList(["X", "I", "iZ", "-iY"])
            value = pauli.compose("X")
            self.assertEqual(target, value)

        with self.subTest(msg="compose single Y"):
            target = PauliList(["Y", "-iZ", "I", "iX"])
            value = pauli.compose("Y")
            self.assertEqual(target, value)

        with self.subTest(msg="compose single Z"):
            target = PauliList(["Z", "iY", "-iX", "I"])
            value = pauli.compose("Z")
            self.assertEqual(target, value)

    def test_dot_1q(self):
        """Test 1-qubit dot method."""
        # Test single qubit Pauli dot products
        pauli = PauliList(["I", "X", "Y", "Z"])

        with self.subTest(msg="dot single I"):
            target = PauliList(["I", "X", "Y", "Z"])
            value = pauli.dot("I")
            self.assertEqual(target, value)

        with self.subTest(msg="dot single X"):
            target = PauliList(["X", "I", "-iZ", "iY"])
            value = pauli.dot("X")
            self.assertEqual(target, value)

        with self.subTest(msg="dot single Y"):
            target = PauliList(["Y", "iZ", "I", "-iX"])
            value = pauli.dot("Y")
            self.assertEqual(target, value)

        with self.subTest(msg="dot single Z"):
            target = PauliList(["Z", "-iY", "iX", "I"])
            value = pauli.dot("Z")
            self.assertEqual(target, value)

    def test_qargs_compose_1q(self):
        """Test 1-qubit compose method with qargs."""

        pauli1 = PauliList(["III", "XXX"])
        pauli2 = PauliList("Z")

        with self.subTest(msg="compose 1-qubit qargs=[0]"):
            target = PauliList(["IIZ", "iXXY"])
            value = pauli1.compose(pauli2, qargs=[0])
            self.assertEqual(value, target)

        with self.subTest(msg="compose 1-qubit qargs=[1]"):
            target = PauliList(["IZI", "iXYX"])
            value = pauli1.compose(pauli2, qargs=[1])
            self.assertEqual(value, target)

        with self.subTest(msg="compose 1-qubit qargs=[2]"):
            target = PauliList(["ZII", "iYXX"])
            value = pauli1.compose(pauli2, qargs=[2])
            self.assertEqual(value, target)

    def test_qargs_dot_1q(self):
        """Test 1-qubit dot method with qargs."""

        pauli1 = PauliList(["III", "XXX"])
        pauli2 = PauliList("Z")

        with self.subTest(msg="dot 1-qubit qargs=[0]"):
            target = PauliList(["IIZ", "-iXXY"])
            value = pauli1.dot(pauli2, qargs=[0])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 1-qubit qargs=[1]"):
            target = PauliList(["IZI", "-iXYX"])
            value = pauli1.dot(pauli2, qargs=[1])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 1-qubit qargs=[2]"):
            target = PauliList(["ZII", "-iYXX"])
            value = pauli1.dot(pauli2, qargs=[2])
            self.assertEqual(value, target)

    def test_qargs_compose_2q(self):
        """Test 2-qubit compose method with qargs."""

        pauli1 = PauliList(["III", "XXX"])
        pauli2 = PauliList("ZY")

        with self.subTest(msg="compose 2-qubit qargs=[0, 1]"):
            target = PauliList(["IZY", "XYZ"])
            value = pauli1.compose(pauli2, qargs=[0, 1])
            self.assertEqual(value, target)

        with self.subTest(msg="compose 2-qubit qargs=[1, 0]"):
            target = PauliList(["IYZ", "XZY"])
            value = pauli1.compose(pauli2, qargs=[1, 0])
            self.assertEqual(value, target)

        with self.subTest(msg="compose 2-qubit qargs=[0, 2]"):
            target = PauliList(["ZIY", "YXZ"])
            value = pauli1.compose(pauli2, qargs=[0, 2])
            self.assertEqual(value, target)

        with self.subTest(msg="compose 2-qubit qargs=[2, 0]"):
            target = PauliList(["YIZ", "ZXY"])
            value = pauli1.compose(pauli2, qargs=[2, 0])
            self.assertEqual(value, target)

    def test_qargs_dot_2q(self):
        """Test 2-qubit dot method with qargs."""

        pauli1 = PauliList(["III", "XXX"])
        pauli2 = PauliList("ZY")

        with self.subTest(msg="dot 2-qubit qargs=[0, 1]"):
            target = PauliList(["IZY", "XYZ"])
            value = pauli1.dot(pauli2, qargs=[0, 1])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 2-qubit qargs=[1, 0]"):
            target = PauliList(["IYZ", "XZY"])
            value = pauli1.dot(pauli2, qargs=[1, 0])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 2-qubit qargs=[0, 2]"):
            target = PauliList(["ZIY", "YXZ"])
            value = pauli1.dot(pauli2, qargs=[0, 2])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 2-qubit qargs=[2, 0]"):
            target = PauliList(["YIZ", "ZXY"])
            value = pauli1.dot(pauli2, qargs=[2, 0])
            self.assertEqual(value, target)

    def test_qargs_compose_3q(self):
        """Test 3-qubit compose method with qargs."""

        pauli1 = PauliList(["III", "XXX"])
        pauli2 = PauliList("XYZ")

        with self.subTest(msg="compose 3-qubit qargs=None"):
            target = PauliList(["XYZ", "IZY"])
            value = pauli1.compose(pauli2)
            self.assertEqual(value, target)

        with self.subTest(msg="compose 3-qubit qargs=[0, 1, 2]"):
            target = PauliList(["XYZ", "IZY"])
            value = pauli1.compose(pauli2, qargs=[0, 1, 2])
            self.assertEqual(value, target)

        with self.subTest(msg="compose 3-qubit qargs=[2, 1, 0]"):
            target = PauliList(["ZYX", "YZI"])
            value = pauli1.compose(pauli2, qargs=[2, 1, 0])
            self.assertEqual(value, target)

        with self.subTest(msg="compose 3-qubit qargs=[1, 0, 2]"):
            target = PauliList(["XZY", "IYZ"])
            value = pauli1.compose(pauli2, qargs=[1, 0, 2])
            self.assertEqual(value, target)

    def test_qargs_dot_3q(self):
        """Test 3-qubit dot method with qargs."""

        pauli1 = PauliList(["III", "XXX"])
        pauli2 = PauliList("XYZ")

        with self.subTest(msg="dot 3-qubit qargs=None"):
            target = PauliList(["XYZ", "IZY"])
            value = pauli1.dot(pauli2, qargs=[0, 1, 2])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 3-qubit qargs=[0, 1, 2]"):
            target = PauliList(["XYZ", "IZY"])
            value = pauli1.dot(pauli2, qargs=[0, 1, 2])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 3-qubit qargs=[2, 1, 0]"):
            target = PauliList(["ZYX", "YZI"])
            value = pauli1.dot(pauli2, qargs=[2, 1, 0])
            self.assertEqual(value, target)

        with self.subTest(msg="dot 3-qubit qargs=[1, 0, 2]"):
            target = PauliList(["XZY", "IYZ"])
            value = pauli1.dot(pauli2, qargs=[1, 0, 2])
            self.assertEqual(value, target)


@ddt
class TestPauliListMethods(QiskitTestCase):
    """Tests for PauliList utility methods class."""

    def test_sort(self):
        """Test sort method."""
        with self.subTest(msg="1 qubit standard order"):
            unsrt = ["X", "Z", "I", "Y", "-iI", "X", "Z", "iI", "-I", "-iY"]
            srt = ["I", "-iI", "-I", "iI", "X", "X", "Y", "-iY", "Z", "Z"]
            target = PauliList(srt)
            value = PauliList(unsrt).sort()
            self.assertEqual(target, value)

        with self.subTest(msg="1 qubit weight order"):
            unsrt = ["X", "Z", "I", "Y", "-iI", "X", "Z", "iI", "-I", "-iY"]
            srt = ["I", "-iI", "-I", "iI", "X", "X", "Y", "-iY", "Z", "Z"]
            target = PauliList(srt)
            value = PauliList(unsrt).sort(weight=True)
            self.assertEqual(target, value)

        with self.subTest(msg="1 qubit phase order"):
            unsrt = ["X", "Z", "I", "Y", "-iI", "X", "Z", "iI", "-I", "-iY"]
            srt = ["I", "X", "X", "Y", "Z", "Z", "-iI", "-iY", "-I", "iI"]
            target = PauliList(srt)
            value = PauliList(unsrt).sort(phase=True)
            self.assertEqual(target, value)

        with self.subTest(msg="1 qubit weight & phase order"):
            unsrt = ["X", "Z", "I", "Y", "-iI", "X", "Z", "iI", "-I", "-iY"]
            srt = ["I", "X", "X", "Y", "Z", "Z", "-iI", "-iY", "-I", "iI"]
            target = PauliList(srt)
            value = PauliList(unsrt).sort(weight=True, phase=True)
            self.assertEqual(target, value)

        with self.subTest(msg="2 qubit standard order"):
            srt = [
                "II",
                "IX",
                "IX",
                "IY",
                "IZ",
                "iIZ",
                "XI",
                "XX",
                "XX",
                "iXX",
                "XY",
                "XZ",
                "iXZ",
                "YI",
                "YI",
                "-YI",
                "YX",
                "-iYX",
                "YY",
                "-iYY",
                "-YY",
                "iYY",
                "YZ",
                "ZI",
                "ZX",
                "ZX",
                "ZY",
                "ZZ",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort()
            self.assertEqual(target, value)

        with self.subTest(msg="2 qubit weight order"):
            srt = [
                "II",
                "IX",
                "IX",
                "IY",
                "IZ",
                "iIZ",
                "XI",
                "YI",
                "YI",
                "-YI",
                "ZI",
                "XX",
                "XX",
                "iXX",
                "XY",
                "XZ",
                "iXZ",
                "YX",
                "-iYX",
                "YY",
                "YY",
                "-YY",
                "YZ",
                "ZX",
                "ZX",
                "ZY",
                "ZZ",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort(weight=True)
            self.assertEqual(target, value)

        with self.subTest(msg="2 qubit phase order"):
            srt = [
                "II",
                "IX",
                "IX",
                "IY",
                "IZ",
                "XI",
                "XX",
                "XX",
                "XY",
                "XZ",
                "YI",
                "YI",
                "YX",
                "YY",
                "YY",
                "YZ",
                "ZI",
                "ZX",
                "ZX",
                "ZY",
                "ZZ",
                "-iYX",
                "-YI",
                "-YY",
                "iIZ",
                "iXX",
                "iXZ",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort(phase=True)
            self.assertEqual(target, value)

        with self.subTest(msg="2 qubit weight & phase order"):
            srt = [
                "II",
                "IX",
                "IX",
                "IY",
                "IZ",
                "XI",
                "YI",
                "YI",
                "ZI",
                "XX",
                "XX",
                "XY",
                "XZ",
                "YX",
                "YY",
                "YY",
                "YZ",
                "ZX",
                "ZX",
                "ZY",
                "ZZ",
                "-iYX",
                "-YI",
                "-YY",
                "iIZ",
                "iXX",
                "iXZ",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort(weight=True, phase=True)
            self.assertEqual(target, value)

        with self.subTest(msg="3 qubit standard order"):
            srt = [
                "III",
                "III",
                "IIX",
                "IIY",
                "-IIY",
                "IIZ",
                "IXI",
                "IXX",
                "IXY",
                "iIXY",
                "IXZ",
                "IYI",
                "IYX",
                "IYY",
                "IYZ",
                "IZI",
                "IZX",
                "IZY",
                "IZY",
                "IZZ",
                "XII",
                "XII",
                "XIX",
                "XIY",
                "XIZ",
                "XXI",
                "XXX",
                "-iXXX",
                "XXY",
                "XXZ",
                "XYI",
                "XYX",
                "iXYX",
                "XYY",
                "XYZ",
                "XYZ",
                "XZI",
                "XZX",
                "XZY",
                "XZZ",
                "YII",
                "YIX",
                "YIY",
                "YIZ",
                "YXI",
                "YXX",
                "YXY",
                "YXZ",
                "YXZ",
                "YYI",
                "YYX",
                "YYX",
                "YYY",
                "YYZ",
                "YZI",
                "YZX",
                "YZY",
                "YZZ",
                "ZII",
                "ZIX",
                "ZIY",
                "ZIZ",
                "ZXI",
                "ZXX",
                "iZXX",
                "ZXY",
                "ZXZ",
                "ZYI",
                "ZYI",
                "ZYX",
                "ZYY",
                "ZYZ",
                "ZZI",
                "ZZX",
                "ZZY",
                "ZZZ",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort()
            self.assertEqual(target, value)

        with self.subTest(msg="3 qubit weight order"):
            srt = [
                "III",
                "III",
                "IIX",
                "IIY",
                "-IIY",
                "IIZ",
                "IXI",
                "IYI",
                "IZI",
                "XII",
                "XII",
                "YII",
                "ZII",
                "IXX",
                "IXY",
                "iIXY",
                "IXZ",
                "IYX",
                "IYY",
                "IYZ",
                "IZX",
                "IZY",
                "IZY",
                "IZZ",
                "XIX",
                "XIY",
                "XIZ",
                "XXI",
                "XYI",
                "XZI",
                "YIX",
                "YIY",
                "YIZ",
                "YXI",
                "YYI",
                "YZI",
                "ZIX",
                "ZIY",
                "ZIZ",
                "ZXI",
                "ZYI",
                "ZYI",
                "ZZI",
                "XXX",
                "-iXXX",
                "XXY",
                "XXZ",
                "XYX",
                "iXYX",
                "XYY",
                "XYZ",
                "XYZ",
                "XZX",
                "XZY",
                "XZZ",
                "YXX",
                "YXY",
                "YXZ",
                "YXZ",
                "YYX",
                "YYX",
                "YYY",
                "YYZ",
                "YZX",
                "YZY",
                "YZZ",
                "ZXX",
                "iZXX",
                "ZXY",
                "ZXZ",
                "ZYX",
                "ZYY",
                "ZYZ",
                "ZZX",
                "ZZY",
                "ZZZ",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort(weight=True)
            self.assertEqual(target, value)

        with self.subTest(msg="3 qubit phase order"):
            srt = [
                "III",
                "III",
                "IIX",
                "IIY",
                "IIZ",
                "IXI",
                "IXX",
                "IXY",
                "IXZ",
                "IYI",
                "IYX",
                "IYY",
                "IYZ",
                "IZI",
                "IZX",
                "IZY",
                "IZY",
                "IZZ",
                "XII",
                "XII",
                "XIX",
                "XIY",
                "XIZ",
                "XXI",
                "XXX",
                "XXY",
                "XXZ",
                "XYI",
                "XYX",
                "XYY",
                "XYZ",
                "XYZ",
                "XZI",
                "XZX",
                "XZY",
                "XZZ",
                "YII",
                "YIX",
                "YIY",
                "YIZ",
                "YXI",
                "YXX",
                "YXY",
                "YXZ",
                "YXZ",
                "YYI",
                "YYX",
                "YYX",
                "YYY",
                "YYZ",
                "YZI",
                "YZX",
                "YZY",
                "YZZ",
                "ZII",
                "ZIX",
                "ZIY",
                "ZIZ",
                "ZXI",
                "ZXX",
                "ZXY",
                "ZXZ",
                "ZYI",
                "ZYI",
                "ZYX",
                "ZYY",
                "ZYZ",
                "ZZI",
                "ZZX",
                "ZZY",
                "ZZZ",
                "-iXXX",
                "-IIY",
                "iIXY",
                "iXYX",
                "iZXX",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort(phase=True)
            self.assertEqual(target, value)

        with self.subTest(msg="3 qubit weight & phase order"):
            srt = [
                "III",
                "III",
                "IIX",
                "IIY",
                "IYI",
                "IZI",
                "XII",
                "XII",
                "YII",
                "ZII",
                "IXX",
                "IXY",
                "IXZ",
                "IYX",
                "IYY",
                "IYZ",
                "IZX",
                "IZY",
                "IZY",
                "IZZ",
                "XIX",
                "XIY",
                "XIZ",
                "XXI",
                "XYI",
                "XZI",
                "YIX",
                "YIY",
                "YIZ",
                "YYI",
                "YZI",
                "ZIX",
                "ZIY",
                "ZIZ",
                "ZXI",
                "ZYI",
                "ZYI",
                "ZZI",
                "XXX",
                "XXY",
                "XXZ",
                "XYX",
                "XYY",
                "XYZ",
                "XZX",
                "XZY",
                "XZZ",
                "YXX",
                "YXY",
                "YXZ",
                "YXZ",
                "YYX",
                "YYX",
                "YYY",
                "YYZ",
                "YZX",
                "YZY",
                "YZZ",
                "ZXX",
                "ZXY",
                "ZXZ",
                "ZYX",
                "ZYY",
                "ZYZ",
                "ZZX",
                "ZZY",
                "ZZZ",
                "-iZIZ",
                "-iXXX",
                "-IIY",
                "iIXI",
                "iIXY",
                "iYXI",
                "iXYX",
                "iXYZ",
                "iZXX",
            ]
            unsrt = srt.copy()
            np.random.shuffle(unsrt)
            target = PauliList(srt)
            value = PauliList(unsrt).sort(weight=True, phase=True)
            self.assertEqual(target, value)

    def test_unique(self):
        """Test unique method."""
        with self.subTest(msg="1 qubit"):
            labels = ["X", "Z", "X", "X", "I", "Y", "I", "X", "Z", "Z", "X", "I"]
            unique = ["X", "Z", "I", "Y"]
            target = PauliList(unique)
            value = PauliList(labels).unique()
            self.assertEqual(target, value)

        with self.subTest(msg="2 qubit"):
            labels = ["XX", "IX", "XX", "II", "IZ", "ZI", "YX", "YX", "ZZ", "IX", "XI"]
            unique = ["XX", "IX", "II", "IZ", "ZI", "YX", "ZZ", "XI"]
            target = PauliList(unique)
            value = PauliList(labels).unique()
            self.assertEqual(target, value)

        with self.subTest(msg="10 qubit"):
            labels = [10 * "X", 10 * "I", 10 * "X"]
            unique = [10 * "X", 10 * "I"]
            target = PauliList(unique)
            value = PauliList(labels).unique()
            self.assertEqual(target, value)

    def test_delete(self):
        """Test delete method."""
        with self.subTest(msg="no rows"):
            pauli = PauliList(["XX", "ZZ"])
            self.assertEqual(pauli.delete([]), pauli)

        with self.subTest(msg="single row"):
            for j in range(1, 6):
                pauli = PauliList([j * "X", j * "Y"])
                self.assertEqual(pauli.delete(0), PauliList(j * "Y"))
                self.assertEqual(pauli.delete(1), PauliList(j * "X"))

        with self.subTest(msg="multiple rows"):
            for j in range(1, 6):
                pauli = PauliList([j * "X", "-i" + j * "Y", j * "Z"])
                self.assertEqual(pauli.delete([0, 2]), PauliList("-i" + j * "Y"))
                self.assertEqual(pauli.delete([1, 2]), PauliList(j * "X"))
                self.assertEqual(pauli.delete([0, 1]), PauliList(j * "Z"))

        with self.subTest(msg="no qubits"):
            pauli = PauliList(["XX", "ZZ"])
            self.assertEqual(pauli.delete([], qubit=True), pauli)

        with self.subTest(msg="single qubit"):
            pauli = PauliList(["IIX", "iIYI", "ZII"])
            value = pauli.delete(0, qubit=True)
            target = PauliList(["II", "iIY", "ZI"])
            self.assertEqual(value, target)
            value = pauli.delete(1, qubit=True)
            target = PauliList(["IX", "iII", "ZI"])
            self.assertEqual(value, target)
            value = pauli.delete(2, qubit=True)
            target = PauliList(["IX", "iYI", "II"])
            self.assertEqual(value, target)

        with self.subTest(msg="multiple qubits"):
            pauli = PauliList(["IIX", "IYI", "-ZII"])
            value = pauli.delete([0, 1], qubit=True)
            target = PauliList(["I", "I", "-Z"])
            self.assertEqual(value, target)
            value = pauli.delete([1, 2], qubit=True)
            target = PauliList(["X", "I", "-I"])
            self.assertEqual(value, target)
            value = pauli.delete([0, 2], qubit=True)
            target = PauliList(["I", "Y", "-I"])
            self.assertEqual(value, target)

    def test_insert(self):
        """Test insert method."""
        # Insert single row
        for j in range(1, 10):
            pauli = PauliList(j * "X")
            target0 = PauliList([j * "I", j * "X"])
            target1 = PauliList([j * "X", j * "I"])

            with self.subTest(msg=f"single row from str ({j})"):
                value0 = pauli.insert(0, j * "I")
                self.assertEqual(value0, target0)
                value1 = pauli.insert(1, j * "I")
                self.assertEqual(value1, target1)

            with self.subTest(msg=f"single row from PauliList ({j})"):
                value0 = pauli.insert(0, PauliList(j * "I"))
                self.assertEqual(value0, target0)
                value1 = pauli.insert(1, PauliList(j * "I"))
                self.assertEqual(value1, target1)

            target0 = PauliList(["i" + j * "I", j * "X"])
            target1 = PauliList([j * "X", "i" + j * "I"])

            with self.subTest(msg=f"single row with phase from str ({j})"):
                value0 = pauli.insert(0, "i" + j * "I")
                self.assertEqual(value0, target0)
                value1 = pauli.insert(1, "i" + j * "I")
                self.assertEqual(value1, target1)

            with self.subTest(msg=f"single row with phase from PauliList ({j})"):
                value0 = pauli.insert(0, PauliList("i" + j * "I"))
                self.assertEqual(value0, target0)
                value1 = pauli.insert(1, PauliList("i" + j * "I"))
                self.assertEqual(value1, target1)

        # Insert multiple rows
        for j in range(1, 10):
            pauli = PauliList("i" + j * "X")
            insert = PauliList([j * "I", j * "Y", j * "Z", "-i" + j * "X"])
            target0 = insert + pauli
            target1 = pauli + insert

            with self.subTest(msg=f"multiple-rows from PauliList ({j})"):
                value0 = pauli.insert(0, insert)
                self.assertEqual(value0, target0)
                value1 = pauli.insert(1, insert)
                self.assertEqual(value1, target1)

        # Insert single column
        pauli = PauliList(["X", "Y", "Z", "-iI"])
        for i in ["I", "X", "Y", "Z", "iY"]:
            phase = "" if len(i) == 1 else i[0]
            p = i if len(i) == 1 else i[1]
            target0 = PauliList(
                [
                    phase + "X" + p,
                    phase + "Y" + p,
                    phase + "Z" + p,
                    ("" if phase else "-i") + "I" + p,
                ]
            )
            target1 = PauliList(
                [
                    i + "X",
                    i + "Y",
                    i + "Z",
                    ("" if phase else "-i") + p + "I",
                ]
            )

            with self.subTest(msg="single-column single-val from str"):
                value = pauli.insert(0, i, qubit=True)
                self.assertEqual(value, target0)
                value = pauli.insert(1, i, qubit=True)
                self.assertEqual(value, target1)

            with self.subTest(msg="single-column single-val from PauliList"):
                value = pauli.insert(0, PauliList(i), qubit=True)
                self.assertEqual(value, target0)
                value = pauli.insert(1, PauliList(i), qubit=True)
                self.assertEqual(value, target1)

        # Insert single column with multiple values
        pauli = PauliList(["X", "Y", "iZ"])
        for i in [["I", "X", "Y"], ["X", "iY", "Z"], ["Y", "Z", "I"]]:
            target0 = PauliList(
                ["X" + i[0], "Y" + i[1] if len(i[1]) == 1 else i[1][0] + "Y" + i[1][1], "iZ" + i[2]]
            )
            target1 = PauliList([i[0] + "X", i[1] + "Y", "i" + i[2] + "Z"])

            with self.subTest(msg="single-column multiple-vals from PauliList"):
                value = pauli.insert(0, PauliList(i), qubit=True)
                self.assertEqual(value, target0)
                value = pauli.insert(1, PauliList(i), qubit=True)
                self.assertEqual(value, target1)

        # Insert multiple columns from single
        pauli = PauliList(["X", "iY", "Z"])
        for j in range(1, 5):
            for i in [j * "I", j * "X", j * "Y", "i" + j * "Z"]:
                phase = "" if len(i) == j else i[0]
                p = i if len(i) == j else i[1:]
                target0 = PauliList(
                    [
                        phase + "X" + p,
                        ("-" if phase else "i") + "Y" + p,
                        phase + "Z" + p,
                    ]
                )
                target1 = PauliList([i + "X", ("-" if phase else "i") + p + "Y", i + "Z"])

                with self.subTest(msg="multiple-columns single-val from str"):
                    value = pauli.insert(0, i, qubit=True)
                    self.assertEqual(value, target0)
                    value = pauli.insert(1, i, qubit=True)
                    self.assertEqual(value, target1)

                with self.subTest(msg="multiple-columns single-val from PauliList"):
                    value = pauli.insert(0, PauliList(i), qubit=True)
                    self.assertEqual(value, target0)
                    value = pauli.insert(1, PauliList(i), qubit=True)
                    self.assertEqual(value, target1)

        # Insert multiple columns multiple row values
        pauli = PauliList(["X", "Y", "-iZ"])
        for j in range(1, 5):
            for i in [
                [j * "I", j * "X", j * "Y"],
                [j * "X", j * "Z", "i" + j * "Y"],
                [j * "Y", j * "Z", j * "I"],
            ]:
                target0 = PauliList(
                    [
                        "X" + i[0],
                        "Y" + i[1],
                        ("-i" if len(i[2]) == j else "") + "Z" + i[2][-j:],
                    ]
                )
                target1 = PauliList(
                    [
                        i[0] + "X",
                        i[1] + "Y",
                        ("-i" if len(i[2]) == j else "") + i[2][-j:] + "Z",
                    ]
                )

                with self.subTest(msg="multiple-column multiple-vals from PauliList"):
                    value = pauli.insert(0, PauliList(i), qubit=True)
                    self.assertEqual(value, target0)
                    value = pauli.insert(1, PauliList(i), qubit=True)
                    self.assertEqual(value, target1)

    def test_commutes(self):
        """Test commutes method."""
        # Single qubit Pauli
        pauli = PauliList(["I", "X", "Y", "Z", "-iY"])
        with self.subTest(msg="commutes single-Pauli I"):
            value = list(pauli.commutes("I"))
            target = [True, True, True, True, True]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli X"):
            value = list(pauli.commutes("X"))
            target = [True, True, False, False, False]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli Y"):
            value = list(pauli.commutes("Y"))
            target = [True, False, True, False, True]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli Z"):
            value = list(pauli.commutes("Z"))
            target = [True, False, False, True, False]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli iZ"):
            value = list(pauli.commutes("iZ"))
            target = [True, False, False, True, False]
            self.assertEqual(value, target)

        # 2-qubit Pauli
        pauli = PauliList(["II", "IX", "YI", "XY", "ZZ", "-iYY"])
        with self.subTest(msg="commutes single-Pauli II"):
            value = list(pauli.commutes("II"))
            target = [True, True, True, True, True, True]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli IX"):
            value = list(pauli.commutes("IX"))
            target = [True, True, True, False, False, False]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli XI"):
            value = list(pauli.commutes("XI"))
            target = [True, True, False, True, False, False]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli YI"):
            value = list(pauli.commutes("YI"))
            target = [True, True, True, False, False, True]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli IY"):
            value = list(pauli.commutes("IY"))
            target = [True, False, True, True, False, True]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli XY"):
            value = list(pauli.commutes("XY"))
            target = [True, False, False, True, True, False]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli YX"):
            value = list(pauli.commutes("YX"))
            target = [True, True, True, True, True, False]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli ZZ"):
            value = list(pauli.commutes("ZZ"))
            target = [True, False, False, True, True, True]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes single-Pauli iYX"):
            value = list(pauli.commutes("iYX"))
            target = [True, True, True, True, True, False]
            self.assertEqual(value, target)

    def test_anticommutes(self):
        """Test anticommutes method."""
        # Single qubit Pauli
        pauli = PauliList(["I", "X", "Y", "Z", "-iY"])
        with self.subTest(msg="anticommutes single-Pauli I"):
            value = list(pauli.anticommutes("I"))
            target = [False, False, False, False, False]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli X"):
            value = list(pauli.anticommutes("X"))
            target = [False, False, True, True, True]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli Y"):
            value = list(pauli.anticommutes("Y"))
            target = [False, True, False, True, False]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli Z"):
            value = list(pauli.anticommutes("Z"))
            target = [False, True, True, False, True]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli iZ"):
            value = list(pauli.anticommutes("iZ"))
            target = [False, True, True, False, True]
            self.assertEqual(value, target)

        # 2-qubit Pauli
        pauli = PauliList(["II", "IX", "YI", "XY", "ZZ", "iZX"])
        with self.subTest(msg="anticommutes single-Pauli II"):
            value = list(pauli.anticommutes("II"))
            target = [False, False, False, False, False, False]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli IX"):
            value = list(pauli.anticommutes("IX"))
            target = [False, False, False, True, True, False]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli XI"):
            value = list(pauli.anticommutes("XI"))
            target = [False, False, True, False, True, True]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli YI"):
            value = list(pauli.anticommutes("YI"))
            target = [False, False, False, True, True, True]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli IY"):
            value = list(pauli.anticommutes("IY"))
            target = [False, True, False, False, True, True]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli XY"):
            value = list(pauli.anticommutes("XY"))
            target = [False, True, True, False, False, False]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli YX"):
            value = list(pauli.anticommutes("YX"))
            target = [False, False, False, False, False, True]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli ZZ"):
            value = list(pauli.anticommutes("ZZ"))
            target = [False, True, True, False, False, True]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes single-Pauli iXY"):
            value = list(pauli.anticommutes("iXY"))
            target = [False, True, True, False, False, False]
            self.assertEqual(value, target)

    def test_commutes_with_all(self):
        """Test commutes_with_all method."""
        # 1-qubit
        pauli = PauliList(["I", "X", "Y", "Z", "-iY"])
        with self.subTest(msg="commutes_with_all [I]"):
            value = list(pauli.commutes_with_all("I"))
            target = [0, 1, 2, 3, 4]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [X]"):
            value = list(pauli.commutes_with_all("X"))
            target = [0, 1]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [Y]"):
            value = list(pauli.commutes_with_all("Y"))
            target = [0, 2, 4]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [Z]"):
            value = list(pauli.commutes_with_all("Z"))
            target = [0, 3]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [iY]"):
            value = list(pauli.commutes_with_all("iY"))
            target = [0, 2, 4]
            self.assertEqual(value, target)

        # 2-qubit Pauli
        pauli = PauliList(["II", "IX", "YI", "XY", "ZZ", "iXY"])

        with self.subTest(msg="commutes_with_all [IX, YI]"):
            other = PauliList(["IX", "YI"])
            value = list(pauli.commutes_with_all(other))
            target = [0, 1, 2]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [XY, ZZ]"):
            other = PauliList(["XY", "ZZ"])
            value = list(pauli.commutes_with_all(other))
            target = [0, 3, 4, 5]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [YX, ZZ]"):
            other = PauliList(["YX", "ZZ"])
            value = list(pauli.commutes_with_all(other))
            target = [0, 3, 4, 5]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [XY, YX]"):
            other = PauliList(["XY", "YX"])
            value = list(pauli.commutes_with_all(other))
            target = [0, 3, 4, 5]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [XY, IX]"):
            other = PauliList(["XY", "IX"])
            value = list(pauli.commutes_with_all(other))
            target = [0]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [YX, IX]"):
            other = PauliList(["YX", "IX"])
            value = list(pauli.commutes_with_all(other))
            target = [0, 1, 2]
            self.assertEqual(value, target)

        with self.subTest(msg="commutes_with_all [-iYX, iZZ]"):
            other = PauliList(["-iYX", "iZZ"])
            value = list(pauli.commutes_with_all(other))
            target = [0, 3, 4, 5]
            self.assertEqual(value, target)

    def test_anticommutes_with_all(self):
        """Test anticommutes_with_all method."""
        # 1-qubit
        pauli = PauliList(["I", "X", "Y", "Z", "-iY"])
        with self.subTest(msg="anticommutes_with_all [I]"):
            value = list(pauli.anticommutes_with_all("I"))
            target = []
            self.assertEqual(value, target)

        with self.subTest(msg="antianticommutes_with_all [X]"):
            value = list(pauli.anticommutes_with_all("X"))
            target = [2, 3, 4]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [Y]"):
            value = list(pauli.anticommutes_with_all("Y"))
            target = [1, 3]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [Z]"):
            value = list(pauli.anticommutes_with_all("Z"))
            target = [1, 2, 4]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [iY]"):
            value = list(pauli.anticommutes_with_all("iY"))
            target = [1, 3]
            self.assertEqual(value, target)

        # 2-qubit Pauli
        pauli = PauliList(["II", "IX", "YI", "XY", "ZZ", "iZX"])

        with self.subTest(msg="anticommutes_with_all [IX, YI]"):
            other = PauliList(["IX", "YI"])
            value = list(pauli.anticommutes_with_all(other))
            target = [3, 4]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [XY, ZZ]"):
            other = PauliList(["XY", "ZZ"])
            value = list(pauli.anticommutes_with_all(other))
            target = [1, 2]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [YX, ZZ]"):
            other = PauliList(["YX", "ZZ"])
            value = list(pauli.anticommutes_with_all(other))
            target = [5]
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [XY, YX]"):
            other = PauliList(["XY", "YX"])
            value = list(pauli.anticommutes_with_all(other))
            target = []
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [XY, IX]"):
            other = PauliList(["XY", "IX"])
            value = list(pauli.anticommutes_with_all(other))
            target = []
            self.assertEqual(value, target)

        with self.subTest(msg="anticommutes_with_all [YX, IX]"):
            other = PauliList(["YX", "IX"])
            value = list(pauli.anticommutes_with_all(other))
            target = []
            self.assertEqual(value, target)

    @combine(
        gate=(
            IGate(),
            XGate(),
            YGate(),
            ZGate(),
            HGate(),
            SGate(),
            SdgGate(),
            Clifford(IGate()),
            Clifford(XGate()),
            Clifford(YGate()),
            Clifford(ZGate()),
            Clifford(HGate()),
            Clifford(SGate()),
            Clifford(SdgGate()),
        )
    )
    def test_evolve_clifford1(self, gate):
        """Test evolve method for 1-qubit Clifford gates."""
        op = Operator(gate)
        pauli_list = PauliList(pauli_group_labels(1, True))
        value = [Operator(pauli) for pauli in pauli_list.evolve(gate)]
        value_h = [Operator(pauli) for pauli in pauli_list.evolve(gate, frame="h")]
        value_s = [Operator(pauli) for pauli in pauli_list.evolve(gate, frame="s")]
        if isinstance(gate, Clifford):
            value_inv = [Operator(pauli) for pauli in pauli_list.evolve(gate.adjoint())]
        else:
            value_inv = [Operator(pauli) for pauli in pauli_list.evolve(gate.inverse())]
        target = [op.adjoint().dot(pauli).dot(op) for pauli in pauli_list]
        self.assertListEqual(value, target)
        self.assertListEqual(value, value_h)
        self.assertListEqual(value_inv, value_s)

    @combine(
        gate=(
            CXGate(),
            CYGate(),
            CZGate(),
            SwapGate(),
            Clifford(CXGate()),
            Clifford(CYGate()),
            Clifford(CZGate()),
            Clifford(SwapGate()),
        )
    )
    def test_evolve_clifford2(self, gate):
        """Test evolve method for 2-qubit Clifford gates."""
        op = Operator(gate)
        pauli_list = PauliList(pauli_group_labels(2, True))
        value = [Operator(pauli) for pauli in pauli_list.evolve(gate)]
        value_h = [Operator(pauli) for pauli in pauli_list.evolve(gate, frame="h")]
        value_s = [Operator(pauli) for pauli in pauli_list.evolve(gate, frame="s")]
        if isinstance(gate, Clifford):
            value_inv = [Operator(pauli) for pauli in pauli_list.evolve(gate.adjoint())]
        else:
            value_inv = [Operator(pauli) for pauli in pauli_list.evolve(gate.inverse())]
        target = [op.adjoint().dot(pauli).dot(op) for pauli in pauli_list]
        self.assertListEqual(value, target)
        self.assertListEqual(value, value_h)
        self.assertListEqual(value_inv, value_s)

    def test_phase_dtype_evolve_clifford(self):
        """Test phase dtype during evolve method for Clifford gates."""
        gates = (
            IGate(),
            XGate(),
            YGate(),
            ZGate(),
            HGate(),
            SGate(),
            SdgGate(),
            CXGate(),
            CYGate(),
            CZGate(),
            SwapGate(),
        )
        dtypes = [
            int,
            np.int8,
            np.uint8,
            np.int16,
            np.uint16,
            np.int32,
            np.uint32,
            np.int64,
            np.uint64,
        ]
        for gate, dtype in itertools.product(gates, dtypes):
            z = np.ones(gate.num_qubits, dtype=bool)
            x = np.ones(gate.num_qubits, dtype=bool)
            phase = (np.sum(z & x) % 4).astype(dtype)
            paulis = Pauli((z, x, phase))
            evo = paulis.evolve(gate)
            self.assertEqual(evo.phase.dtype, dtype)

    @combine(phase=(True, False))
    def test_evolve_clifford_qargs(self, phase):
        """Test evolve method for random Clifford"""
        cliff = random_clifford(3, seed=10)
        op = Operator(cliff)
        pauli_list = random_pauli_list(5, 3, seed=10, phase=phase)
        qargs = [3, 0, 1]
        value = [Operator(pauli) for pauli in pauli_list.evolve(cliff, qargs=qargs)]
        value_inv = [Operator(pauli) for pauli in pauli_list.evolve(cliff.adjoint(), qargs=qargs)]
        value_h = [Operator(pauli) for pauli in pauli_list.evolve(cliff, qargs=qargs, frame="h")]
        value_s = [Operator(pauli) for pauli in pauli_list.evolve(cliff, qargs=qargs, frame="s")]
        target = [
            Operator(pauli).compose(op.adjoint(), qargs=qargs).dot(op, qargs=qargs)
            for pauli in pauli_list
        ]
        self.assertListEqual(value, target)
        self.assertListEqual(value, value_h)
        self.assertListEqual(value_inv, value_s)

    def test_group_qubit_wise_commuting(self):
        """Test grouping qubit-wise commuting operators"""

        def qubitwise_commutes(left: Pauli, right: Pauli) -> bool:
            return len(left) == len(right) and all(a.commutes(b) for a, b in zip(left, right))

        input_labels = ["IY", "ZX", "XZ", "YI", "YX", "YY", "YZ", "ZI", "ZX", "ZY", "iZZ", "II"]
        np.random.shuffle(input_labels)
        pauli_list = PauliList(input_labels)
        groups = pauli_list.group_qubit_wise_commuting()

        # checking that every input Pauli in pauli_list is in a group in the ouput
        output_labels = [pauli.to_label() for group in groups for pauli in group]
        self.assertListEqual(sorted(output_labels), sorted(input_labels))

        # Within each group, every operator qubit-wise commutes with every other operator.
        for group in groups:
            self.assertTrue(
                all(
                    qubitwise_commutes(pauli1, pauli2)
                    for pauli1, pauli2 in itertools.combinations(group, 2)
                )
            )
        # For every pair of groups, at least one element from one does not qubit-wise commute with
        # at least one element of the other.
        for group1, group2 in itertools.combinations(groups, 2):
            self.assertFalse(
                all(
                    qubitwise_commutes(group1_pauli, group2_pauli)
                    for group1_pauli, group2_pauli in itertools.product(group1, group2)
                )
            )

    def test_group_commuting(self):
        """Test general grouping commuting operators"""

        def commutes(left: Pauli, right: Pauli) -> bool:
            return len(left) == len(right) and left.commutes(right)

        input_labels = ["IY", "ZX", "XZ", "YI", "YX", "YY", "YZ", "ZI", "ZX", "ZY", "iZZ", "II"]
        np.random.shuffle(input_labels)
        pauli_list = PauliList(input_labels)
        #  if qubit_wise=True, equivalent to test_group_qubit_wise_commuting
        groups = pauli_list.group_commuting(qubit_wise=False)

        # checking that every input Pauli in pauli_list is in a group in the ouput
        output_labels = [pauli.to_label() for group in groups for pauli in group]
        self.assertListEqual(sorted(output_labels), sorted(input_labels))
        # Within each group, every operator commutes with every other operator.
        for group in groups:
            self.assertTrue(
                all(commutes(pauli1, pauli2) for pauli1, pauli2 in itertools.combinations(group, 2))
            )
        # For every pair of groups, at least one element from one group does not commute with
        # at least one element of the other.
        for group1, group2 in itertools.combinations(groups, 2):
            self.assertFalse(
                all(
                    commutes(group1_pauli, group2_pauli)
                    for group1_pauli, group2_pauli in itertools.product(group1, group2)
                )
            )


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Abstract base class for Quantum Channels.
"""

from __future__ import annotations
import copy
import sys
from abc import abstractmethod
from numbers import Number, Integral

import numpy as np

from qiskit.circuit.instruction import Instruction
from qiskit.exceptions import QiskitError
from qiskit.quantum_info.operators.linear_op import LinearOp
from qiskit.quantum_info.operators.op_shape import OpShape
from qiskit.quantum_info.operators.operator import Operator
from qiskit.quantum_info.operators.predicates import is_positive_semidefinite_matrix
from qiskit.quantum_info.operators.channel.transformations import _transform_rep
from qiskit.quantum_info.operators.channel.transformations import _to_choi
from qiskit.quantum_info.operators.channel.transformations import _to_kraus
from qiskit.quantum_info.operators.channel.transformations import _to_operator
from qiskit.quantum_info.operators.scalar_op import ScalarOp

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class QuantumChannel(LinearOp):
    """Quantum channel representation base class."""

    def __init__(
        self,
        data: list | np.ndarray,
        num_qubits: int | None = None,
        op_shape: OpShape | None = None,
    ):
        """Initialize a quantum channel Superoperator operator.

        Args:
            data (array or list): quantum channel data array.
            op_shape (OpShape): the operator shape of the channel.
            num_qubits (int): the number of qubits if the channel is N-qubit.

        Raises:
            QiskitError: if arguments are invalid.
        """
        self._data = data
        super().__init__(num_qubits=num_qubits, op_shape=op_shape)

    def __repr__(self):
        prefix = f"{self._channel_rep}("
        pad = len(prefix) * " "
        return "{}{},\n{}input_dims={}, output_dims={})".format(
            prefix,
            np.array2string(np.asarray(self.data), separator=", ", prefix=prefix),
            pad,
            self.input_dims(),
            self.output_dims(),
        )

    def __eq__(self, other: Self):
        """Test if two QuantumChannels are equal."""
        if not super().__eq__(other):
            return False
        return np.allclose(self.data, other.data, rtol=self.rtol, atol=self.atol)

    @property
    def data(self):
        """Return data."""
        return self._data

    @property
    def _channel_rep(self):
        """Return channel representation string"""
        return type(self).__name__

    @property
    def settings(self):
        """Return settings."""
        return {
            "data": self.data,
            "input_dims": self.input_dims(),
            "output_dims": self.output_dims(),
        }

    # ---------------------------------------------------------------------
    # LinearOp methods
    # ---------------------------------------------------------------------

    @abstractmethod
    def conjugate(self):
        r"""Return the conjugate quantum channel.

        .. note::
            This is equivalent to the matrix complex conjugate in the
            :class:`~qiskit.quantum_info.SuperOp` representation
            ie. for a channel :math:`\mathcal{E}`, the SuperOp of
            the conjugate channel :math:`\overline{{\mathcal{{E}}}}` is
            :math:`S_{\overline{\mathcal{E}^\dagger}} = \overline{S_{\mathcal{E}}}`.
        """

    @abstractmethod
    def transpose(self) -> Self:
        r"""Return the transpose quantum channel.

        .. note::
            This is equivalent to the matrix transpose in the
            :class:`~qiskit.quantum_info.SuperOp` representation,
            ie. for a channel :math:`\mathcal{E}`, the SuperOp of
            the transpose channel :math:`\mathcal{{E}}^T` is
            :math:`S_{mathcal{E}^T} = S_{\mathcal{E}}^T`.
        """

    def adjoint(self) -> Self:
        r"""Return the adjoint quantum channel.

        .. note::
            This is equivalent to the matrix Hermitian conjugate in the
            :class:`~qiskit.quantum_info.SuperOp` representation
            ie. for a channel :math:`\mathcal{E}`, the SuperOp of
            the adjoint channel :math:`\mathcal{{E}}^\dagger` is
            :math:`S_{\mathcal{E}^\dagger} = S_{\mathcal{E}}^\dagger`.
        """
        return self.conjugate().transpose()

    def power(self, n: float) -> Self:
        r"""Return the power of the quantum channel.

        Args:
            n (float): the power exponent.

        Returns:
            CLASS: the channel :math:`\mathcal{{E}} ^n`.

        Raises:
            QiskitError: if the input and output dimensions of the
                         CLASS are not equal.

        .. note::
            For non-positive or non-integer exponents the power is
            defined as the matrix power of the
            :class:`~qiskit.quantum_info.SuperOp` representation
            ie. for a channel :math:`\mathcal{{E}}`, the SuperOp of
            the powered channel :math:`\mathcal{{E}}^\n` is
            :math:`S_{{\mathcal{{E}}^n}} = S_{{\mathcal{{E}}}}^n`.
        """
        if n > 0 and isinstance(n, Integral):
            return super().power(n)

        # Conversion to superoperator
        if self._input_dim != self._output_dim:
            raise QiskitError("Can only take power with input_dim = output_dim.")
        rep = self._channel_rep
        input_dim, output_dim = self.dim
        superop = _transform_rep(rep, "SuperOp", self._data, input_dim, output_dim)
        superop = np.linalg.matrix_power(superop, n)

        # Convert back to original representation
        ret = copy.copy(self)
        ret._data = _transform_rep("SuperOp", rep, superop, input_dim, output_dim)
        return ret

    def __sub__(self, other) -> Self:
        qargs = getattr(other, "qargs", None)
        if not isinstance(other, type(self)):
            other = type(self)(other)
        return self._add(-other, qargs=qargs)

    def _add(self, other, qargs=None):
        # NOTE: this method must be overridden for subclasses
        # that don't have a linear matrix representation
        # ie Kraus and Stinespring
        if not isinstance(other, type(self)):
            other = type(self)(other)
        self._op_shape._validate_add(other._op_shape, qargs)
        other = ScalarOp._pad_with_identity(self, other, qargs)
        ret = copy.copy(self)
        ret._data = self._data + other._data
        return ret

    def _multiply(self, other):
        # NOTE: this method must be overridden for subclasses
        # that don't have a linear matrix representation
        # ie Kraus and Stinespring
        if not isinstance(other, Number):
            raise QiskitError("other is not a number")
        ret = copy.copy(self)
        ret._data = other * self._data
        return ret

    # ---------------------------------------------------------------------
    # Additional methods
    # ---------------------------------------------------------------------

    def is_cptp(self, atol: float | None = None, rtol: float | None = None) -> bool:
        """Return True if completely-positive trace-preserving (CPTP)."""
        choi = _to_choi(self._channel_rep, self._data, *self.dim)
        return self._is_cp_helper(choi, atol, rtol) and self._is_tp_helper(choi, atol, rtol)

    def is_tp(self, atol: float | None = None, rtol: float | None = None) -> bool:
        """Test if a channel is trace-preserving (TP)"""
        choi = _to_choi(self._channel_rep, self._data, *self.dim)
        return self._is_tp_helper(choi, atol, rtol)

    def is_cp(self, atol: float | None = None, rtol: float | None = None) -> bool:
        """Test if Choi-matrix is completely-positive (CP)"""
        choi = _to_choi(self._channel_rep, self._data, *self.dim)
        return self._is_cp_helper(choi, atol, rtol)

    def is_unitary(self, atol: float | None = None, rtol: float | None = None) -> bool:
        """Return True if QuantumChannel is a unitary channel."""
        try:
            op = self.to_operator()
            return op.is_unitary(atol=atol, rtol=rtol)
        except QiskitError:
            return False

    def to_operator(self) -> Operator:
        """Try to convert channel to a unitary representation Operator."""
        mat = _to_operator(self._channel_rep, self._data, *self.dim)
        return Operator(mat, self.input_dims(), self.output_dims())

    def to_instruction(self) -> Instruction:
        """Convert to a Kraus or UnitaryGate circuit instruction.

        If the channel is unitary it will be added as a unitary gate,
        otherwise it will be added as a kraus simulator instruction.

        Returns:
            qiskit.circuit.Instruction: A kraus instruction for the channel.

        Raises:
            QiskitError: if input data is not an N-qubit CPTP quantum channel.
        """

        # Check if input is an N-qubit CPTP channel.
        num_qubits = int(np.log2(self._input_dim))
        if self._input_dim != self._output_dim or 2**num_qubits != self._input_dim:
            raise QiskitError(
                "Cannot convert QuantumChannel to Instruction: channel is not an N-qubit channel."
            )
        if not self.is_cptp():
            raise QiskitError("Cannot convert QuantumChannel to Instruction: channel is not CPTP.")
        # Next we convert to the Kraus representation. Since channel is CPTP we know
        # that there is only a single set of Kraus operators
        kraus, _ = _to_kraus(self._channel_rep, self._data, *self.dim)
        # If we only have a single Kraus operator then the channel is
        # a unitary channel so can be converted to a UnitaryGate. We do this by
        # converting to an Operator and using its to_instruction method
        if len(kraus) == 1:
            return Operator(kraus[0]).to_instruction()
        return Instruction("kraus", num_qubits, 0, kraus)

    def _is_cp_helper(self, choi, atol, rtol):
        """Test if a channel is completely-positive (CP)"""
        if atol is None:
            atol = self.atol
        if rtol is None:
            rtol = self.rtol
        return is_positive_semidefinite_matrix(choi, rtol=rtol, atol=atol)

    def _is_tp_helper(self, choi, atol, rtol):
        """Test if Choi-matrix is trace-preserving (TP)"""
        if atol is None:
            atol = self.atol
        if rtol is None:
            rtol = self.rtol
        # Check if the partial trace is the identity matrix
        d_in, d_out = self.dim
        mat = np.trace(np.reshape(choi, (d_in, d_out, d_in, d_out)), axis1=1, axis2=3)
        tp_cond = np.linalg.eigvalsh(mat - np.eye(len(mat)))
        zero = np.isclose(tp_cond, 0, atol=atol, rtol=rtol)
        return np.all(zero)

    def _format_state(self, state, density_matrix=False):
        """Format input state so it is statevector or density matrix"""
        state = np.array(state)
        shape = state.shape
        ndim = state.ndim
        if ndim > 2:
            raise QiskitError("Input state is not a vector or matrix.")
        # Flatten column-vector to vector
        if ndim == 2:
            if shape[1] != 1 and shape[1] != shape[0]:
                raise QiskitError("Input state is not a vector or matrix.")
            if shape[1] == 1:
                # flatten column-vector to vector
                state = np.reshape(state, shape[0])
        # Convert statevector to density matrix if required
        if density_matrix and ndim == 1:
            state = np.outer(state, np.transpose(np.conj(state)))
        return state

    @abstractmethod
    def _evolve(self, state, qargs=None):
        """Evolve a quantum state by the quantum channel.

        Args:
            state (DensityMatrix or Statevector): The input state.
            qargs (list): a list of quantum state subsystem positions to apply
                           the quantum channel on.

        Returns:
            DensityMatrix: the output quantum state as a density matrix.

        Raises:
            QiskitError: if the quantum channel dimension does not match the
                         specified quantum state subsystem dimensions.
        """
        pass

    @classmethod
    def _init_transformer(cls, data):
        """Convert input into a QuantumChannel subclass object or Operator object"""
        # This handles common conversion for all QuantumChannel subclasses.
        # If the input is already a QuantumChannel subclass it will return
        # the original object
        if isinstance(data, QuantumChannel):
            return data
        if hasattr(data, "to_quantumchannel"):
            # If the data object is not a QuantumChannel it will give
            # preference to a 'to_quantumchannel' attribute that allows
            # an arbitrary object to define its own conversion to any
            # quantum channel subclass.
            return data.to_quantumchannel()
        if hasattr(data, "to_channel"):
            # TODO: this 'to_channel' method is the same case as the above
            # but is used by current version of Aer. It should be removed
            # once Aer is nupdated to use `to_quantumchannel`
            # instead of `to_channel`,
            return data.to_channel()
        # Finally if the input is not a QuantumChannel and doesn't have a
        # 'to_quantumchannel' conversion method we try and initialize it as a
        # regular matrix Operator which can be converted into a QuantumChannel.
        return Operator(data)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Replace each sequence of Clifford gates by a single Clifford gate."""

from functools import partial

from qiskit.transpiler.passes.optimization.collect_and_collapse import (
    CollectAndCollapse,
    collect_using_filter_function,
    collapse_to_operation,
)

from qiskit.quantum_info.operators import Clifford


class CollectCliffords(CollectAndCollapse):
    """Collects blocks of Clifford gates and replaces them by a :class:`~qiskit.quantum_info.Clifford`
    object.
    """

    def __init__(
        self,
        do_commutative_analysis=False,
        split_blocks=True,
        min_block_size=2,
        split_layers=False,
        collect_from_back=False,
    ):
        """CollectCliffords initializer.

        Args:
            do_commutative_analysis (bool): if True, exploits commutativity relations
                between nodes.
            split_blocks (bool): if True, splits collected blocks into sub-blocks
                over disjoint qubit subsets.
            min_block_size (int): specifies the minimum number of gates in the block
                for the block to be collected.
            split_layers (bool): if True, splits collected blocks into sub-blocks
                over disjoint qubit subsets.
            collect_from_back (bool): specifies if blocks should be collected started
                from the end of the circuit.
        """

        collect_function = partial(
            collect_using_filter_function,
            filter_function=_is_clifford_gate,
            split_blocks=split_blocks,
            min_block_size=min_block_size,
            split_layers=split_layers,
            collect_from_back=collect_from_back,
        )
        collapse_function = partial(collapse_to_operation, collapse_function=_collapse_to_clifford)

        super().__init__(
            collect_function=collect_function,
            collapse_function=collapse_function,
            do_commutative_analysis=do_commutative_analysis,
        )


clifford_gate_names = [
    "x",
    "y",
    "z",
    "h",
    "s",
    "sdg",
    "cx",
    "cy",
    "cz",
    "swap",
    "clifford",
    "linear_function",
    "pauli",
]


def _is_clifford_gate(node):
    """Specifies whether a node holds a clifford gate."""
    return node.op.name in clifford_gate_names and getattr(node.op, "condition", None) is None


def _collapse_to_clifford(circuit):
    """Specifies how to construct a ``Clifford`` from a quantum circuit (that must
    consist of Clifford gates only)."""
    return Clifford(circuit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Estimator class
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any

import numpy as np

from qiskit.circuit import QuantumCircuit
from qiskit.exceptions import QiskitError
from qiskit.quantum_info import Statevector
from qiskit.quantum_info.operators.base_operator import BaseOperator

from .base import BaseEstimator, EstimatorResult
from .primitive_job import PrimitiveJob
from .utils import (
    _circuit_key,
    _observable_key,
    bound_circuit_to_instruction,
    init_observable,
)


class Estimator(BaseEstimator[PrimitiveJob[EstimatorResult]]):
    """
    Reference implementation of :class:`BaseEstimator`.

    :Run Options:

        - **shots** (None or int) --
          The number of shots. If None, it calculates the exact expectation
          values. Otherwise, it samples from normal distributions with standard errors as standard
          deviations using normal distribution approximation.

        - **seed** (np.random.Generator or int) --
          Set a fixed seed or generator for the normal distribution. If shots is None,
          this option is ignored.
    """

    def __init__(self, *, options: dict | None = None):
        """
        Args:
            options: Default options.

        Raises:
            QiskitError: if some classical bits are not used for measurements.
        """
        super().__init__(options=options)
        self._circuits = []
        self._parameters = []
        self._observables = []
        self._circuit_ids = {}
        self._observable_ids = {}

    def _call(
        self,
        circuits: Sequence[int],
        observables: Sequence[int],
        parameter_values: Sequence[Sequence[float]],
        **run_options,
    ) -> EstimatorResult:
        shots = run_options.pop("shots", None)
        seed = run_options.pop("seed", None)
        if seed is None:
            rng = np.random.default_rng()
        elif isinstance(seed, np.random.Generator):
            rng = seed
        else:
            rng = np.random.default_rng(seed)

        # Initialize metadata
        metadata: list[dict[str, Any]] = [{} for _ in range(len(circuits))]

        bound_circuits = []
        for i, value in zip(circuits, parameter_values):
            if len(value) != len(self._parameters[i]):
                raise QiskitError(
                    f"The number of values ({len(value)}) does not match "
                    f"the number of parameters ({len(self._parameters[i])})."
                )
            bound_circuits.append(
                self._circuits[i]
                if len(value) == 0
                else self._circuits[i].assign_parameters(dict(zip(self._parameters[i], value)))
            )
        sorted_observables = [self._observables[i] for i in observables]
        expectation_values = []
        for circ, obs, metadatum in zip(bound_circuits, sorted_observables, metadata):
            if circ.num_qubits != obs.num_qubits:
                raise QiskitError(
                    f"The number of qubits of a circuit ({circ.num_qubits}) does not match "
                    f"the number of qubits of a observable ({obs.num_qubits})."
                )
            final_state = Statevector(bound_circuit_to_instruction(circ))
            expectation_value = final_state.expectation_value(obs)
            if shots is None:
                expectation_values.append(expectation_value)
            else:
                expectation_value = np.real_if_close(expectation_value)
                sq_obs = (obs @ obs).simplify(atol=0)
                sq_exp_val = np.real_if_close(final_state.expectation_value(sq_obs))
                variance = sq_exp_val - expectation_value**2
                variance = max(variance, 0)
                standard_error = np.sqrt(variance / shots)
                expectation_value_with_error = rng.normal(expectation_value, standard_error)
                expectation_values.append(expectation_value_with_error)
                metadatum["variance"] = variance
                metadatum["shots"] = shots

        return EstimatorResult(np.real_if_close(expectation_values), metadata)

    def _run(
        self,
        circuits: tuple[QuantumCircuit, ...],
        observables: tuple[BaseOperator, ...],
        parameter_values: tuple[tuple[float, ...], ...],
        **run_options,
    ):
        circuit_indices = []
        for circuit in circuits:
            key = _circuit_key(circuit)
            index = self._circuit_ids.get(key)
            if index is not None:
                circuit_indices.append(index)
            else:
                circuit_indices.append(len(self._circuits))
                self._circuit_ids[key] = len(self._circuits)
                self._circuits.append(circuit)
                self._parameters.append(circuit.parameters)
        observable_indices = []
        for observable in observables:
            observable = init_observable(observable)
            index = self._observable_ids.get(_observable_key(observable))
            if index is not None:
                observable_indices.append(index)
            else:
                observable_indices.append(len(self._observables))
                self._observable_ids[_observable_key(observable)] = len(self._observables)
                self._observables.append(observable)
        job = PrimitiveJob(
            self._call, circuit_indices, observable_indices, parameter_values, **run_options
        )
        job._submit()
        return job



#!/usr/bin/env python3
# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Compare Qiskit versions to determine if we should run qpy compat tests."""

import argparse
import sys

from qiskit.qpy.interface import VERSION_PATTERN_REGEX


def main():
    """Main function."""
    parser = argparse.ArgumentParser(prog="compare_version", description="Compare version strings")
    parser.add_argument(
        "source_version", help="Source version of Qiskit that is generating the payload"
    )
    parser.add_argument("test_version", help="Version under test that will load the QPY payload")
    args = parser.parse_args()
    source_match = VERSION_PATTERN_REGEX.search(args.source_version)
    target_match = VERSION_PATTERN_REGEX.search(args.test_version)
    source_version = tuple(int(x) for x in source_match.group("release").split("."))
    target_version = tuple(int(x) for x in target_match.group("release").split("."))
    if source_version > target_version:
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import copy
import pickle
import ddt

from qiskit.circuit import ClassicalRegister, Clbit
from qiskit.circuit.classical import expr, types
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt.ddt
class TestExprProperties(QiskitTestCase):
    def test_bool_type_is_singleton(self):
        """The `Bool` type is meant (and used) as a Python singleton object for efficiency.  It must
        always be referentially equal to all other references to it."""
        self.assertIs(types.Bool(), types.Bool())
        self.assertIs(types.Bool(), copy.copy(types.Bool()))
        self.assertIs(types.Bool(), copy.deepcopy(types.Bool()))
        self.assertIs(types.Bool(), pickle.loads(pickle.dumps(types.Bool())))

    @ddt.data(types.Bool(), types.Uint(8))
    def test_types_can_be_cloned(self, obj):
        """Test that various ways of cloning a `Type` object are valid and produce equal output."""
        self.assertEqual(obj, copy.copy(obj))
        self.assertEqual(obj, copy.deepcopy(obj))
        self.assertEqual(obj, pickle.loads(pickle.dumps(obj)))

    @ddt.data(
        expr.Var(ClassicalRegister(3, "c"), types.Uint(3)),
        expr.Value(3, types.Uint(2)),
        expr.Cast(expr.Value(1, types.Uint(8)), types.Bool()),
        expr.Unary(expr.Unary.Op.LOGIC_NOT, expr.Value(False, types.Bool()), types.Bool()),
        expr.Binary(
            expr.Binary.Op.LOGIC_OR,
            expr.Value(False, types.Bool()),
            expr.Value(True, types.Bool()),
            types.Bool(),
        ),
    )
    def test_expr_can_be_cloned(self, obj):
        """Test that various ways of cloning an `Expr` object are valid and produce equal output."""
        self.assertEqual(obj, copy.copy(obj))
        self.assertEqual(obj, copy.deepcopy(obj))
        self.assertEqual(obj, pickle.loads(pickle.dumps(obj)))

    def test_var_standalone(self):
        """Test that the ``Var.standalone`` property is set correctly."""
        self.assertFalse(expr.Var(Clbit(), types.Bool()).standalone)
        self.assertFalse(expr.Var(ClassicalRegister(8, "cr"), types.Uint(8)).standalone)

    def test_var_hashable(self):
        clbits = [Clbit(), Clbit()]
        cregs = [ClassicalRegister(2, "cr1"), ClassicalRegister(2, "cr2")]

        vars_ = [
            expr.Var(clbits[0], types.Bool()),
            expr.Var(clbits[1], types.Bool()),
            expr.Var(cregs[0], types.Uint(2)),
            expr.Var(cregs[1], types.Uint(2)),
        ]
        duplicates = [
            expr.Var(clbits[0], types.Bool()),
            expr.Var(clbits[1], types.Bool()),
            expr.Var(cregs[0], types.Uint(2)),
            expr.Var(cregs[1], types.Uint(2)),
        ]

        # Smoke test.
        self.assertEqual(vars_, duplicates)
        # Actual test of hashability properties.
        self.assertEqual(set(vars_ + duplicates), set(vars_))
